#!/usr/bin/env python
# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-
'''a yum plugin to choose a stable, working, up-to-date mirror

To install this plugin, drop it into /usr/lib/yum-plugins/, drop
stablemirror.conf into /etc/yum/pluginconf.d/, and make sure you have
'plugins=1' in your /etc/yum.conf.

Configuration Options:
----------------------
/etc/yum/pluginconf.d/stablemirror.conf:
    [main]
    enabled=1
    stablemirrorspath=/var/cache/yum/stablemirrors
    # 4*60*60*24, or 4 days:
    stuck=345600

License:
--------
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

Theory:
-------
Yum normally uses a randomly chosen mirror each time it is run.  When a
repository is updated, its mirrors update at various times, some more
often than others.  While the mirrors are catching up, some will be in a
group that is up-to-date and otherw will be in groups at old dates. 
There are more such groups if the repository is updated more often. 
Some mirrors break and get stuck, often for months.  Thus, a randomly
chosen mirror may be more or less up-to-date than the mirror used last
time.  If less up-to-date, then updated packages will be removed from
the lists.  If an error occurs during downloading, and yum switches to
another mirror, that mirror may not be at the same date as the previous
mirror, and its lists and packages can have the wrong checksums.

Successful updating needs the first mirror to be at least as up-to-date
as the mirror last used, and, when failing over to another mirror, the
new mirror must at the same date as the first mirror.  The various
mirrors should be used evenly, but this doesn't need to be done each
time run; once per machine is enough.  Stablemirror:

    * starts with the same mirror each time
    * when changing mirrors on an error, makes sure that the new mirror
      has the same date as the first mirror
    * ensures that the first mirror is at least as up-to-date as last
      time yum was run
    * avoids mirrors the user dislikes
    * never uses mirrors the user rejects
    * but notices new mirrors as they become available
    * provides a way out:  Ctl-C prompts for user action:
    
        - quit (if -y option used, quits without prompting)
        - continue with the next mirror
        - avoid the previous mirror
        - reject the previous mirror

In the future, stablemirror may skip known broken mirrors, checking them
only occasionally.

The list of stable mirrors is kept in the yum cache, and (with the next
version of yum) may be removed by "yum clean all".  It is easy to edit
if there are mirrors that you want to avoid or reject that don't give
enough time to Ctl-C them.  Newly added mirrors are inserted randomly
into the stable list (which might introduce some slight "instability").

The yum ui is not rich enough to support the Ctl-C interface, having no
means to prompt for general input.  Any graphical yum that supports
plugins (?) will need to use the -y option to prevent the Ctl-C handler
from printing invisible input requests on the command line.  Sorry.

(C) Copyright 2006, 2007 George A. Nelson  tonynelson@georgeanelson.com


Changelog:

* Sun 11 Jun 2006 Tony Nelson <tonynelson@georgeanelson.com> - 0.1
- initial release

* Tue 27 Jun 2006 Tony Nelson <tonynelson@georgeanelson.com> - 0.2
- Ctl-C reliable handling

* Fri 23 Feb 2007 Tony Nelson <tonynelson@georgeanelson.com> - 0.2.1
- updated for yum 3.0 in FC6

* Sun 25 Feb 2007 Tony Nelson <tonynelson@georgeanelson.com> - 0.2.2
- interrupt_callback.moveto() safe when no mirror set
- interrupt_callback() simpler help / huh? parsing, indented help output.

* Wed 17 Oct 2007 Tony Nelson <tonynelson@georgeanelson.com> - 0.4.0
- updated for yum 3.2 in F7.

'''

__author__  = 'Tony Nelson'
__email__   = 'tonynelson@georgeanelson.com'
__url__     = 'http://georgeanelson.com/stablemirror.htm'
__copyright__ = 'Copyright 2006, 2007 by George A. Nelson.  All rights reserved.'
__license__ = 'GPL'
__date__    = '17 October 2007'
__version__ = '0.4.0'

# yum.__init__.doRepoSetup() fetches repoMD before postreposetup_hook()
# (and after prereposetup_hook()).  To always affect the mirror
# ordering, do it in our MirrorGroup subclass.  In prereposetup_hook(),
# read our stablemirror file, set up our MirrorGroup subclass, patch
# yum.repos.Repository.setupGrab() to make our MirrorGroup subclass and
# install the KeyboardInterrupt callback, and patch
# yum.repos.Repository._checkRepoXML() to check dates.  Eww. 
# Yum.repos.setupGrab() should have a way to specify a MirrorGroup
# subclass, rather than just the hardwired two.
#
# I think fastestmirror got it wrong, or just isn't gross enough, as it
# doesn't affect mirror order until after the repoMD fetch, which
# defaults to using a random mirror which may already be known to
# fastestmirror as a bad mirror.

# We could detect broken mirrors when fetching repoMD, either as the
# initial fetch or when switching mirrors on a failure.  Deem permanently
# broken if the repoMD is not found, and move to reject list (though it
# may be Fedora's fault for listing a mirror for repos it doesn't do,
# even if the mirror does do some of the repos it is listed for, but what
# can one do other than bugzilla?).  Deem temporarily broken if the
# repoMD is sufficiently out-of-date (>4 days?), and don't try again
# until 1/2 the amount out-of-date.  Deem temporarily broken if the
# repoMD is otherwise inaccessable at that mirror, record the initial
# time, don't try again for 1/2 the time since then (hopefully
# consistently non-responding mirrors such as technoera.com will quickly
# stop being used).  Keep track of the number of times a mirror works
# and the number of times it fails, and deem temporarily broken a mirror
# that fails more 1/2 the time, and handle as for inaccessible mirrors. 
# Each mirror needs its base url, the time of detection, the time to try
# again, the success count, and the fail count (?).
#
# Umm, note that auto-reject of bad mirrors isn't necessary.  We work
# properly as-is.  We might be nicer and faster and less chatty if we know
# how to auto-reject.

# we ought to find the newest repoXML date, to avoid latching on to a stuck
# mirror.  We can't just use the current date, as some repos really don't
# change.  If we knew the "master" repo, we could ask it, but if we're using
# mirrors the baseurl isn't available, and there might not be a unique master
# anyway.  We should, when there is no repomd.xml in our_setupGrap(), poll all
# the available mirrors (stable and avoid) to find the most current date.  Use
# the threaded flash fetching from reposinsync / fastestmirror.


import os
import sys
import time
from urlparse import urlparse
import types
import random
import tempfile
from yum.yumRepo import YumRepository
from urlgrabber.mirror import MirrorGroup
from urlgrabber.grabber import URLGrabError, CallbackObject
#from repomd import repoMDObject, mdErrors
from yum import repoMDObject
from yum.Errors import RepoMDError
import signal

import yum.Errors
from yum.plugins import TYPE_INTERACTIVE, TYPE_CORE
from yum.plugins import PluginYumExit


requires_api_version = '2.2'
plugin_type = (TYPE_INTERACTIVE, TYPE_CORE)

us = 'stablemirror'

stuck = 0
stablemirrorspath = ''
stablemirrorssections = ['stable', 'avoid', 'reject']
stablemirrors = []          # a mirror is an object with named properties, including base_url
avoidmirrors = []
rejectmirrors = []
stablemirrorssectionslist = [(s, globals()[s+'mirrors']) for s in stablemirrorssections]
changed = False
assumeyes = False
interrupted_mirror = None


class MirrorData:
    '''Hold info about each mirror, including base_url
    '''
    pass

def debug(msg):
    #print msg
    pass


def init_hook(conduit):
    global stablemirrorspath, stuck, assumeyes
    stablemirrorspath = conduit.confString( 'main', 'stablemirrorspath',
        default='/var/cache/yum/stablemirrors' )
    stuck = conduit.confInt( 'main', 'stuck', default=4*60*60*24 )
    assumeyes = conduit.getConf().assumeyes

def prereposetup_hook(conduit):
    conduit.info(2, '%s: the easily edited stablemirror file is "%s"' % (us, stablemirrorspath))
    read_stablemirrors(conduit)

def close_hook(conduit):
    debug('close_hook entering')
    if changed:
        write_stablemirrors(conduit)
    debug('close_hook leaving')

def clean_hook(conduit):    # api_version 2.3+
    if os.path.exists(stablemirrorspath):
        os.unlink(stablemirrorspath)


def read_stablemirrors(conduit):
    debug('read_stablemirrors entering')
    sections = dict(stablemirrorssectionslist)
    try:
        f = file(stablemirrorspath)
    except IOError:
        conduit.info(2, '%s: no stablemirror file' % us)
    else:
        try:
            try:
                cur_list = None
                for line in f:
                    data = line.lstrip()
                    if not len(data) or data[0] == '#':
                        pass                            # blank line or comment
                    elif len(line) == len(data):
                        data = data.strip()             # not indented: new section
                        if not data in sections:
                            conduit.error(2, '%s: unknown section "%s" at:\n  "%s"' % (us, data, line))
                            raise PlugInYumExit()
                        cur_list = sections[data]
                    elif cur_list is None:
                        conduit.error(2, '%s: specify a section before:\n  "%s"' % (us, line))
                        raise PlugInYumExit()
                    else:
                        data_list = data.split()
                        md = MirrorData()
                        md.base_url = data_list[0]
                        cur_list.append(md)             # append to current list
            except IOError:
                conduit.error(2, '%s: can\'t read stablemirror file' % us)
                raise PlugInYumExit()
        finally:
            f.close()
    debug('read_stablemirrors leaving')

def write_stablemirrors(conduit):
    debug('write_stablemirrors entering')
    try:
        f = file(stablemirrorspath, 'w')
    except IOError:
        conduit.error(2, '%s: can\'t create stablemirror file "%s"' % (us, stablemirrorspath))
    else:
        try:
            try:
                f.write('# yum stablemirror plugin\n')
                f.write('# list mirror sites in order to use, avoid until last resort, or reject entirely\n')
                f.write('# mirrors are not ordered or grouped by repository.\n')
                def write_section(name, cur_list):
                    debug(' -- name=%s  cur_list=%s' % (name, [m.base_url for m in cur_list]))
                    f.write('\n'+name+'\n')
                    for md in cur_list:
                        f.write('\t'+md.base_url+'\n')
                for s, l in stablemirrorssectionslist:
                    write_section(s, l)
            except IOError:
                conduit.error(2, '%s: can\'t write stablemirror file "%s"' % (us, stablemirrorspath))
        finally:
            f.close()
    debug('write_stablemirrors leaving')

def list_stablemirrors(repo_mirrors):
    '''Cull and reorder a repo's mirrors.  Return ordered by stable, new,
    and avoid lists.
    '''
    debug('list_stablemirrors entering repo_mirrors=%s' % repo_mirrors)
    rejectmirrors_dict = dict.fromkeys([m.base_url for m in rejectmirrors])
    repo_mirrors_dict = dict(repo_mirrors)
    for k in rejectmirrors_dict.keys():
        if k in repo_mirrors_dict:
            del repo_mirrors_dict[k]
    def add_mirrors(from_mirrors):
        mirrors = []
        for k in from_mirrors:
            if k.base_url in repo_mirrors_dict:
                mirrors.append(k.base_url)
                del repo_mirrors_dict[k.base_url]
        return mirrors
    sl = add_mirrors(stablemirrors)
    al = add_mirrors(avoidmirrors)
    nl = repo_mirrors_dict.keys()
    if len(nl):
        # Riffle new mirrors into stablemirrors.  Stablemirrors original entries
        # must keep their original order, so don't shuffle.  O(n^2), sigh.
        for m in nl:
            r = random.randint(0, len(stablemirrors))
            md = MirrorData()
            md.base_url = m
            stablemirrors.insert(r,md)
        global changed
        changed = True
    debug('list_stablemirrors leaving sl=%s nl=%s al=%s' % (sl, nl, al))
    return [repo_mirrors[k] for k in sl + nl + al]

def cmp_repoXML(xa, xb):
    '''Compare primary timestamps in xa and xb.  Return the time difference,
    or None if unordered (xa or xb is None).
    '''
    debug('cmp_repoXML entering')
    if not xa or not xb:
        return None
    tsa = int(xa.getData('primary').timestamp)
    tsb = int(xb.getData('primary').timestamp)
    debug('cmp_repoXML leaving diff=%d' % (tsa-tsb,))
    return tsa - tsb


# patch setupGrab to use our mirror class

def our_setupGrab(self):    # Repository
    debug('our_setupGrab entering repo=%s' % self.id)
    # get repoMD used last time
    self.local_repoXML = None
    local = os.path.join(self.basecachedir, self.id, 'repomd.xml')
    debug('our_setupGrab local=%s' % local)
    if os.path.exists(local):
        try:
            self.local_repoXML = repoMDObject.RepoMD(self.id, local)
            debug('our_setupGrab parsed local_repoXML')
        except RepoMDError:
            pass
    if not self.local_repoXML:
        debug('our_setupGrab didn\'t get local_repoXML')
    
    self.timeout = None
    save_setupGrab(self)
    self._grab = MGStableMirror(self._grabfunc, self.urls, self)
    self._grabfunc.opts.interrupt_callback = interrupt_callback
    debug('our_setupGrab leaving')

save_setupGrab = YumRepository._setupGrab
YumRepository._setupGrab = our_setupGrab


# patch _checkRepoXML to ensure up-to-date

def our_checkRepoXML(self, fo):    # Repository
    debug('our_checkRepoXML entering')
    # copied from _checkRepoXML, but saving parse repoXML:
    #
    if type(fo) is types.InstanceType:
        filepath = fo.filename
    else:
        filepath = fo
    
    try:
        repoXML = repoMDObject.RepoMD(self.id, filepath)
    except RepoMDError, e:
        raise URLGrabError(-1, 'Error importing repomd.xml for %s: %s' % (self, e))
    
    # new stuff:
    #
    debug('our_checkRepoXML parsing')
    if not getattr(self, 'local_repoXML', None):
        print(' ***** defaulting local_repoXML *****')
        self.local_repoXML = repoXML
    # new repoMD must be at least as recent as the repoMD from last time run
    td = cmp_repoXML(repoXML, self.local_repoXML)
    if td is None:
        # can't happen
        raise ValueError('Incomparable repoXML=%s local_repoXML=%s' % (repoXML, self.local_repoXML))
    if td < -stuck:
        # repo is stuck, and should be avoided for a while
        raise URLGrabError(512, 'repo stuck in the past')
    if td < 0:
        raise URLGrabError(-1, 'repo not up-to-date')
    debug('our_checkRepoXML leaving')

YumRepository._checkRepoXML = our_checkRepoXML


class MGStableMirror(MirrorGroup):
    '''A mirror group that uses mirrors in a stable order.
    '''
    
    def __init__(self, grabber, mirrors, repo, **kwargs):
        debug('MGStableMirror.__init__ entering mirrors=%s' % mirrors)
        MirrorGroup.__init__(self, grabber, mirrors, **kwargs)
        self.repo = repo
        
        # prune and reorder mirrors list
        repomirrors = {}
        for m in self.mirrors:  # a list of dicts
            url = m['mirror']
            base_url = urlparse(url)[1]
            repomirrors[base_url] = m
        self.mirrors = list_stablemirrors(repomirrors)
        debug('MGStableMirror.__init__ leaving')
    
    def _get_mirror(self, gr):
        debug('_get_mirror entering')
        # return current mirror, but also ensure that it's repoXML matches
        # delete non-matching mirrors from mirrors list
        while 1:
            m = MirrorGroup._get_mirror(self, gr)
            if gr.kw.get('fetching_repoMD'):
                break
            if not m.get('cached_repoXML'):
                debug('_get_mirror fetching repoXML for %s' % repr(m))
                fullurl = self._join_url(m['mirror'], self.repo.repoMDFile)
                #kwargs = dict(m.get('kwargs', {}))
                #kwargs.update(gr.kw)
                #for kw in ['checkfunc', 'limit', 'range', 'filename']:
                #    if kw in kwargs:
                #        del kwargs[kw]          # eww
                kwargs = {'copy_local': 1, 'http_headers': None, 'text': None, 'reget': None}
                f = tempfile.NamedTemporaryFile('w+')
                kwargs['filename'] = f.name
                grabber = m.get('grabber') or self.grabber
                try:
                    fn = grabber.urlgrab(fullurl, **kwargs)
                except URLGrabError, e:
                    pass # should handle as in _failure(); already reported
                    debug('_get_mirror didn\'t get repoXML: %s' % e)
                else:
                    debug('_get_mirror got repoXML')
                    try:
                        m['cached_repoXML'] = repoMDObject.RepoMD(self.repo.id, fn)
                        debug('_get_mirror parsed repoXML')
                    except RepoMDError, e:
                        self.report_failure(e, fullurl)
                f.close()
            td = cmp_repoXML(m.get('cached_repoXML'), self.repo.repoXML)
            if td == 0:
                debug('_get_mirror matched repoXML')
                break
            if td is not None and td < -stuck:
                pass # repo is stuck, and should be avoided for a while
                debug('  -- stuck!')
                self.report_failure(URLGrabError(512, 'repo stuck in the past'), fullurl)
            else:
                self.report_failure(URLGrabError(-1, 'repo not same date'), fullurl)
            self.increment_mirror(gr, {'remove': 1, 'remove_master': 1})
        global interrupted_mirror
        interrupted_mirror = m['mirror']    # eww
        debug('_get_mirror leaving mirror=%s' % repr(m))
        return m
    
    def _failure(self, gr, cb_obj):
        ex = cb_obj.exception
        debug('_failure entering exception=%s' % repr(ex))
        # Rpm sets its own signal handlers, which caused a bug in socket timeouts
        # to eat Ctl-C's, turning them into EWOULDBLOCK,  which gets returned as
        # "... (11, ...".  Note that the underlying socket is a blocking socket,
        # so EWOULDBLOCK is not correct.  With the bug fixed, they become EINTR,
        # which isn't so specific, as it could be any signal, but normally only
        # Ctl-C is expected.  The signal.signal() below should make this code
        # moot.
        if ( hasattr(ex, 'errno') and ex.errno == 4 # IOError
         and (ex.args[1].find(' (11, ') >= 0
           or ex.args[1].find(' (4, ')  >= 0) ):
            try:
                interrupt_callback(None)    # never returns
            except URLGrabError:
                pass                        # but can raise!
        if ( gr.kw.get('fetching_repoMD')
         and hasattr(ex, 'errno') ):
            debug(' -- .errno=%s' % ex.errno)
            if ex.errno in [4, 7, 14, 512]:
                # For certain (?) error codes, mark the mirror as bad
                # 4    - IOError on fetch       
                # 7    - HTTPException
                # 14   - HTTPError (includes .code and .exception attributes)
                # For #14, a .code of 404 should eliminate that mirror, as it
                # doesn't host the repo (put on rejectmirrors list?)
                if ex.errno == 14:
                    debug(' ++ .code=%s' % ex.code)
                if ex.errno == 512:
                    debug(' ++ stuck!')
        MirrorGroup._failure(self, gr, cb_obj)
        debug('_failure leaving')
        
    def _mirror_try(self, func, url, kw):
        debug('_mirror_try entering')
        # do the actual download, looping over the mirrors
        # cache mirror's repomd.xml
        fetching_repoMD = (url == self.repo.repoMDFile and func == 'urlgrab')
        if fetching_repoMD:
            kw = dict(kw)
            kw['fetching_repoMD'] = True
        
        # rpm steals python's SIGINT signal.  Steal it back.  Should be safe, as
        # yum is careful to close the rpm database on exit.
        signal.signal(signal.SIGINT, signal.default_int_handler)
        
        result = MirrorGroup._mirror_try(self, func, url, kw)
        
        global interrupted_mirror
        interrupted_mirror = None   # eww
        
        if fetching_repoMD:
            m = self.mirrors[self._next]
            try:
                m['cached_repoXML'] = repoMDObject.RepoMD(self.repo.id, result)
            except RepoMDError:
                pass
        debug('_mirror_try leaving')
        return result
    
    def report_failure(self, exception, url):
        # derived from urlgrabber.grabber.URLGrabber._retry() and
        # yum.yumRepos.yumRepository.get()
        class FailData:
            pass
        failure_obj = self.repo.failure_obj
        if failure_obj:
            f_func, f_args, f_kwargs = failure_obj
            fd = FailData()
            fd.url = url
            fd.exception = exception
            try:
                f_func(fd, *f_args, **f_kwargs)
            except:
                pass


def interrupt_callback(cbo):
    debug('interrupt_callback entering')
    # I'm pretty sure that I shouldn't use a conduit here.  What I do isn't
    # great either.
    if assumeyes:
        raise KeyboardInterrupt
    
    # Hacked from yum-cli.output.YumOutput.userconfirm.  Note that this only
    # works if yum is using a command-line ui and won't work for any other
    # yum ui.  Fortunately, assumeyes can be used to prevent us from being here.
    def moveto(dest):
        if not interrupted_mirror:
            return
        base_url = urlparse(interrupted_mirror)[1]
        if not base_url:
            return
        def deletefrom(src):
            for i in xrange(len(src)):
                if src[i].base_url == base_url:
                    del src[i]
                    return True;
            return False
        if deletefrom(stablemirrors) or deletefrom(avoidmirrors):
            md = MirrorData()
            md.base_url = base_url
            dest.append(md)
            global changed
            changed = True
    
    print '\n%s!  Mirror="%s".' % (
        (cbo is None) and 'Faked interrupt' or 'Interrupted',
        interrupted_mirror )
    while True:            
        choice = raw_input('Continue, quit, avoid, reject, or help: ')
        choice = choice.lower()
        if 'continue'.startswith(choice):
            raise URLGrabError(15, 'User abort')
        elif 'quit'.startswith(choice):
            raise KeyboardInterrupt
        elif 'avoid'.startswith(choice):
            moveto(avoidmirrors)
            raise URLGrabError(-1, 'avoid mirror')
        elif 'reject'.startswith(choice):
            moveto(rejectmirrors)
            raise URLGrabError(-1, 'reject mirror')
        elif 'help'.startswith(choice):
            print (
                '''    Continue with next mirror\n'''
                '''    Quit yum\n'''
                '''    Avoid the interrupted mirror, so it will be used last\n'''
                '''    Reject the interrupted mirror, so it won\'t be used again''' )
            # loop
        else:
            print 'huh?'
            # loop
